@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013      Mellanox Technologies, Inc.
+ * Copyright (c) 2013-2015 Mellanox Technologies, Inc.
  *                         All rights reserved.
  * Copyright (c) 2014      Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
@@ -19,6 +19,7 @@
 
 #include "oshmem_config.h"
 #include "opal/datatype/opal_convertor.h"
+#include "opal/mca/memchecker/base/base.h"
 #include "orte/include/orte/types.h"
 #include "orte/runtime/orte_globals.h"
 #include "oshmem/mca/spml/ikrit/spml_ikrit.h"
@@ -38,6 +39,19 @@
 #define SPML_IKRIT_PUT_DEBUG    0
 #endif
 
+#define SPML_IKRIT_MXM_POST_SEND(sreq) \
+do { \
+    mxm_error_t err; \
+    err = mxm_req_send(&sreq); \
+    if (MXM_OK != err) { \
+        SPML_ERROR("mxm_req_send (op=%d) failed: %s - aborting", \
+                   sreq.opcode, \
+                   mxm_error_string(err)); \
+        oshmem_shmem_abort(-1); \
+        return OSHMEM_ERROR; \
+    } \
+} while(0)
+
 typedef struct spml_ikrit_am_hdr {
     uint64_t va;
 } spml_ikrit_am_hdr_t;
@@ -85,10 +99,16 @@ static inline mxm_mem_key_t *to_mxm_mkey(sshmem_mkey_t *mkey) {
 }
 #endif
 
+
 static inline void mca_spml_irkit_req_wait(mxm_req_base_t *req)
 {
-    while (!mxm_req_test(req))
+    do {
+        /* do at least one progress since
+         * with some TLs (self, shm) request
+         * can be completed immediately
+         */
         opal_progress();
+    } while (!mxm_req_test(req));
 }
 
 static int mca_spml_ikrit_put_request_free(struct oshmem_request_t** request)
@@ -96,11 +116,12 @@ static int mca_spml_ikrit_put_request_free(struct oshmem_request_t** request)
     mca_spml_ikrit_put_request_t *put_req =
             *(mca_spml_ikrit_put_request_t **) request;
 
-    assert(false == put_req->req_put.req_base.req_free_called);
     OPAL_THREAD_LOCK(&oshmem_request_lock);
+    assert(false == put_req->req_put.req_base.req_free_called);
     put_req->req_put.req_base.req_free_called = true;
     OMPI_FREE_LIST_RETURN_MT( &mca_spml_base_put_requests,
                           (ompi_free_list_item_t*)put_req);
+    opal_memchecker_base_mem_noaccess(put_req, sizeof(*put_req));
     OPAL_THREAD_UNLOCK(&oshmem_request_lock);
 
     *request = SHMEM_REQUEST_NULL; /*MPI_REQUEST_NULL;*/
@@ -144,11 +165,12 @@ static int mca_spml_ikrit_get_request_free(struct oshmem_request_t** request)
     mca_spml_ikrit_get_request_t *get_req =
             *(mca_spml_ikrit_get_request_t **) request;
 
-    assert(false == get_req->req_get.req_base.req_free_called);
     OPAL_THREAD_LOCK(&oshmem_request_lock);
+    assert(false == get_req->req_get.req_base.req_free_called);
     get_req->req_get.req_base.req_free_called = true;
     OMPI_FREE_LIST_RETURN_MT( &mca_spml_base_get_requests,
                           (ompi_free_list_item_t*)get_req);
+    opal_memchecker_base_mem_noaccess(get_req, sizeof(*get_req));
     OPAL_THREAD_UNLOCK(&oshmem_request_lock);
 
     *request = SHMEM_REQUEST_NULL; /*MPI_REQUEST_NULL;*/
@@ -164,7 +186,7 @@ static int mca_spml_ikrit_get_request_cancel(struct oshmem_request_t * request,
 
 static void mca_spml_ikrit_get_request_construct(mca_spml_ikrit_get_request_t* req)
 {
-    req->req_get.req_base.req_type = MCA_SPML_REQUEST_PUT;
+    req->req_get.req_base.req_type = MCA_SPML_REQUEST_GET;
     req->req_get.req_base.req_oshmem.req_free = mca_spml_ikrit_get_request_free;
     req->req_get.req_base.req_oshmem.req_cancel =
             mca_spml_ikrit_get_request_cancel;
@@ -201,6 +223,9 @@ mca_spml_ikrit_t mca_spml_ikrit = {
         mca_spml_base_wait,
         mca_spml_base_wait_nb,
         mca_spml_ikrit_fence,
+        mca_spml_base_rmkey_unpack,
+        mca_spml_base_rmkey_free,
+
         (void*)&mca_spml_ikrit
     }
 };
@@ -239,6 +264,10 @@ static inline mca_spml_ikrit_put_request_t *alloc_put_req(void)
     OMPI_FREE_LIST_WAIT_MT(&mca_spml_base_put_requests, item);
 
     req = (mca_spml_ikrit_put_request_t *) item;
+    opal_memchecker_base_mem_undefined(req, sizeof(*req));
+    opal_memchecker_base_mem_defined(&req->req_put.req_base,
+                                     sizeof(req->req_put.req_base));
+
     req->req_put.req_base.req_free_called = false;
     req->req_put.req_base.req_oshmem.req_complete = false;
 
@@ -253,6 +282,10 @@ static inline mca_spml_ikrit_get_request_t *alloc_get_req(void)
     OMPI_FREE_LIST_WAIT_MT(&mca_spml_base_get_requests, item);
 
     req = (mca_spml_ikrit_get_request_t *) item;
+    opal_memchecker_base_mem_undefined(req, sizeof(*req));
+    opal_memchecker_base_mem_defined(&req->req_get.req_base,
+                                     sizeof(req->req_get.req_base));
+
     req->req_get.req_base.req_free_called = false;
     req->req_get.req_base.req_oshmem.req_complete = false;
 
@@ -342,7 +375,6 @@ OBJ_CLASS_INSTANCE( mxm_peer_t,
 int mca_spml_ikrit_del_procs(oshmem_proc_t** procs, size_t nprocs)
 {
     size_t i, n;
-    opal_list_item_t *item;
     int my_rank = oshmem_my_proc_id();
 
     oshmem_shmem_barrier();
@@ -352,25 +384,21 @@ int mca_spml_ikrit_del_procs(oshmem_proc_t** procs, size_t nprocs)
     }
 #endif
 
-    while (NULL != (item = opal_list_remove_first(&mca_spml_ikrit.active_peers))) {
+    while (NULL != opal_list_remove_first(&mca_spml_ikrit.active_peers)) {
     };
     OBJ_DESTRUCT(&mca_spml_ikrit.active_peers);
 
     for (n = 0; n < nprocs; n++) {
         i = (my_rank + n) % nprocs;
-        if (mca_spml_ikrit.mxm_peers[i]->mxm_conn) {
-            mxm_ep_disconnect(mca_spml_ikrit.mxm_peers[i]->mxm_conn);
-        }
-        if (mca_spml_ikrit.hw_rdma_channel && mca_spml_ikrit.mxm_peers[i]->mxm_hw_rdma_conn) {
+        mxm_ep_disconnect(mca_spml_ikrit.mxm_peers[i]->mxm_conn);
+        if (mca_spml_ikrit.hw_rdma_channel) {
+            assert(mca_spml_ikrit.mxm_peers[i]->mxm_hw_rdma_conn != mca_spml_ikrit.mxm_peers[i]->mxm_conn);
             mxm_ep_disconnect(mca_spml_ikrit.mxm_peers[i]->mxm_hw_rdma_conn);
         }
         destroy_ptl_idx(i);
-        if (mca_spml_ikrit.mxm_peers[i]) {
-            OBJ_RELEASE(mca_spml_ikrit.mxm_peers[i]);
-        }
+        OBJ_RELEASE(mca_spml_ikrit.mxm_peers[i]);
     }
-    if (mca_spml_ikrit.mxm_peers)
-        free(mca_spml_ikrit.mxm_peers);
+    free(mca_spml_ikrit.mxm_peers);
 
     return OSHMEM_SUCCESS;
 }
@@ -379,7 +407,7 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
 {
     spml_ikrit_mxm_ep_conn_info_t *ep_info = NULL;
     spml_ikrit_mxm_ep_conn_info_t *ep_hw_rdma_info = NULL;
-    spml_ikrit_mxm_ep_conn_info_t my_ep_info;
+    spml_ikrit_mxm_ep_conn_info_t my_ep_info = {0,};
 #if MXM_API < MXM_VERSION(2,0)
     mxm_conn_req_t *conn_reqs;
     int timeout;
@@ -402,20 +430,18 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
     }
     memset(conn_reqs, 0x0, sizeof(mxm_conn_req_t));
 #endif
-    ep_info = malloc(nprocs * sizeof(spml_ikrit_mxm_ep_conn_info_t));
+    ep_info = calloc(sizeof(spml_ikrit_mxm_ep_conn_info_t), nprocs);
     if (NULL == ep_info) {
         rc = OSHMEM_ERR_OUT_OF_RESOURCE;
         goto bail;
     }
-    memset(ep_info, 0x0, sizeof(spml_ikrit_mxm_ep_conn_info_t));
 
     if (mca_spml_ikrit.hw_rdma_channel) {
-        ep_hw_rdma_info = malloc(nprocs * sizeof(spml_ikrit_mxm_ep_conn_info_t));
+        ep_hw_rdma_info = calloc(sizeof(spml_ikrit_mxm_ep_conn_info_t), nprocs);
         if (NULL == ep_hw_rdma_info) {
             rc = OSHMEM_ERR_OUT_OF_RESOURCE;
             goto bail;
         }
-        memset(ep_hw_rdma_info, 0x0, sizeof(spml_ikrit_mxm_ep_conn_info_t));
     }
 
     mca_spml_ikrit.mxm_peers = (mxm_peer_t **) malloc(nprocs
@@ -428,11 +454,13 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
 #if MXM_API < MXM_VERSION(2,0)
     if (OSHMEM_SUCCESS
             != spml_ikrit_get_ep_address(&my_ep_info, MXM_PTL_SELF)) {
-        return OSHMEM_ERROR;
+        rc = OSHMEM_ERROR;
+        goto bail;
     }
     if (OSHMEM_SUCCESS
             != spml_ikrit_get_ep_address(&my_ep_info, MXM_PTL_RDMA)) {
-        return OSHMEM_ERROR;
+        rc = OSHMEM_ERROR;
+        goto bail;
     }
 #else
     if (mca_spml_ikrit.hw_rdma_channel) {
@@ -440,7 +468,8 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
         if (MXM_OK != err) {
             orte_show_help("help-oshmem-spml-ikrit.txt", "unable to get endpoint address", true,
                     mxm_error_string(err));
-            return OSHMEM_ERROR;
+            rc = OSHMEM_ERROR;
+            goto bail;
         }
         oshmem_shmem_allgather(&my_ep_info, ep_hw_rdma_info,
                 sizeof(spml_ikrit_mxm_ep_conn_info_t));
@@ -449,7 +478,8 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
     if (MXM_OK != err) {
         orte_show_help("help-oshmem-spml-ikrit.txt", "unable to get endpoint address", true,
                 mxm_error_string(err));
-        return OSHMEM_ERROR;
+        rc = OSHMEM_ERROR;
+        goto bail;
     }
 #endif
     oshmem_shmem_allgather(&my_ep_info, ep_info,
@@ -520,17 +550,18 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
     /* Save returned connections */
     for (i = 0; i < nprocs; ++i) {
         mca_spml_ikrit.mxm_peers[i]->mxm_conn = conn_reqs[i].conn;
-        if (OSHMEM_SUCCESS != create_ptl_idx(i))
+        if (OSHMEM_SUCCESS != create_ptl_idx(i)) {
+            rc = OSHMEM_ERR_CONNECTION_FAILED;
             goto bail;
+        }
 
         mxm_conn_ctx_set(conn_reqs[i].conn, mca_spml_ikrit.mxm_peers[i]);
     }
 
-    if (ep_info)
-        free(ep_info);
-    if (conn_reqs)
-        free(conn_reqs);
+    free(conn_reqs);
 #endif
+    free(ep_info);
+    free(ep_hw_rdma_info);
 
 #if MXM_API >= MXM_VERSION(2,0)
     if (mca_spml_ikrit.bulk_connect) {
@@ -548,7 +579,7 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
             continue;
         }
         if (procs[i] == proc_self)
-        continue;
+            continue;
 
         /* use zcopy for put/get via sysv shared memory */
         procs[i]->transport_ids[0] = MXM_PTL_SHM;
@@ -559,13 +590,12 @@ int mca_spml_ikrit_add_procs(oshmem_proc_t** procs, size_t nprocs)
     SPML_VERBOSE(50, "*** ADDED PROCS ***");
     return OSHMEM_SUCCESS;
 
-    bail:
-    #if MXM_API < MXM_VERSION(2,0)
-        if (conn_reqs)
-            free(conn_reqs);
-        if (ep_info)
-            free(ep_info);
-    #endif
+bail:
+#if MXM_API < MXM_VERSION(2,0)
+    free(conn_reqs);
+#endif
+    free(ep_info);
+    free(ep_hw_rdma_info);
     SPML_ERROR("add procs FAILED rc=%d", rc);
 
     return rc;
@@ -616,7 +646,7 @@ sshmem_mkey_t *mca_spml_ikrit_register(void* addr,
 #if MXM_API < MXM_VERSION(2,0)
             mkeys[i].len = 0;
 #else
-            if (mca_spml_ikrit.ud_only && !mca_spml_ikrit.hw_rdma_channel) {
+            if (mca_spml_ikrit.ud_only) {
                 mkeys[i].len = 0;
                 break;
             }
@@ -643,14 +673,10 @@ sshmem_mkey_t *mca_spml_ikrit_register(void* addr,
             }
 #endif
             break;
-
-        default:
-            SPML_ERROR("unsupported PTL: %d", i);
-            goto error_out;
         }
         SPML_VERBOSE(5,
                      "rank %d ptl %d addr %p size %llu %s",
-                     oshmem_proc_local_proc->proc_name.vpid, i, addr, (unsigned long long)size,
+		     oshmem_proc_pe(oshmem_proc_local()), i, addr, (unsigned long long)size,
                      mca_spml_base_mkey2str(&mkeys[i]));
 
     }
@@ -695,6 +721,8 @@ int mca_spml_ikrit_deregister(sshmem_mkey_t *mkeys)
             break;
         }
     }
+    free(mkeys);
+
     return OSHMEM_SUCCESS;
 
 }
@@ -764,10 +792,7 @@ static int mca_spml_ikrit_get_helper(mxm_send_req_t *sreq,
     /**
      * Get the address to the remote rkey.
      **/
-    r_mkey = mca_memheap.memheap_get_cached_mkey(src,
-                                                 src_addr,
-                                                 ptl_id,
-                                                 &rva);
+    r_mkey = mca_memheap_base_get_cached_mkey(src, src_addr, ptl_id, &rva);
     if (!r_mkey) {
         SPML_ERROR("pe=%d: %p is not address of shared variable",
                    src, src_addr);
@@ -814,10 +839,7 @@ static inline int mca_spml_ikrit_get_shm(void *src_addr,
     if (ptl_id != MXM_PTL_SHM)
         return OSHMEM_ERROR;
 
-    r_mkey = mca_memheap.memheap_get_cached_mkey(src,
-                                                 src_addr,
-                                                 ptl_id,
-                                                 &rva);
+    r_mkey = mca_memheap_base_get_cached_mkey(src, src_addr, ptl_id, &rva);
     if (!r_mkey) {
         SPML_ERROR("pe=%d: %p is not address of shared variable",
                    src, src_addr);
@@ -839,6 +861,7 @@ static inline int mca_spml_ikrit_get_shm(void *src_addr,
 int mca_spml_ikrit_get(void *src_addr, size_t size, void *dst_addr, int src)
 {
     mxm_send_req_t sreq;
+    mxm_error_t err;
 
     if (0 >= size) {
         return OSHMEM_SUCCESS;
@@ -857,17 +880,12 @@ int mca_spml_ikrit_get(void *src_addr, size_t size, void *dst_addr, int src)
         return OSHMEM_ERROR;
     }
 
-#if MXM_API < MXM_VERSION(2,0)
-    sreq.base.flags = MXM_REQ_FLAG_BLOCKING;
-#else
-    sreq.flags = MXM_REQ_SEND_FLAG_BLOCKING;
-#endif
     sreq.base.completed_cb = NULL;
+    sreq.flags = 0;
+
+    SPML_IKRIT_MXM_POST_SEND(sreq);
 
-    mxm_req_send(&sreq);
-    opal_progress();
     mca_spml_irkit_req_wait(&sreq.base);
-
     if (MXM_OK != sreq.base.error) {
         SPML_ERROR("get request failed: %s - aborting",
                    mxm_error_string(sreq.base.error));
@@ -926,14 +944,8 @@ int mca_spml_ikrit_get_async(void *src_addr,
     get_req->mxm_req.base.context = get_req;
     OPAL_THREAD_ADD32(&mca_spml_ikrit.n_active_gets, 1);
 
-    mxm_req_send(&get_req->mxm_req);
+    SPML_IKRIT_MXM_POST_SEND(get_req->mxm_req);
 
-    if (MXM_OK != get_req->mxm_req.base.error) {
-        SPML_ERROR("get request failed: %s - aborting",
-                   mxm_error_string(get_req->mxm_req.base.error));
-        oshmem_shmem_abort(-1);
-        return OSHMEM_ERROR;
-    }
     return OSHMEM_SUCCESS;
 }
 
@@ -980,7 +992,7 @@ static int mca_spml_ikrit_mxm_fence(int dst)
     fence_req->mxm_req.base.context = fence_req;
     OPAL_THREAD_ADD32(&mca_spml_ikrit.n_mxm_fences, 1);
 
-    mxm_req_send(&fence_req->mxm_req);
+    SPML_IKRIT_MXM_POST_SEND(fence_req->mxm_req);
     return OSHMEM_SUCCESS;
 }
 
@@ -1052,10 +1064,7 @@ static inline int mca_spml_ikrit_put_internal(void* dst_addr,
 
     ptl_id = get_ptl_id(dst);
     /* Get rkey of remote PE (dst proc) which must be on memheap  */
-    r_mkey = mca_memheap.memheap_get_cached_mkey(dst,
-                                                 dst_addr,
-                                                 ptl_id,
-                                                 &rva);
+    r_mkey = mca_memheap_base_get_cached_mkey(dst, dst_addr, ptl_id, &rva);
     if (!r_mkey) {
         SPML_ERROR("pe=%d: %p is not address of shared variable",
                    dst, dst_addr);
@@ -1079,10 +1088,7 @@ static inline int mca_spml_ikrit_put_internal(void* dst_addr,
         }
         /* segment not mapped - fallback to rmda */
         ptl_id = MXM_PTL_RDMA;
-        r_mkey = mca_memheap.memheap_get_cached_mkey(dst,
-                                                     dst_addr,
-                                                     ptl_id,
-                                                     &rva);
+        r_mkey = mca_memheap_base_get_cached_mkey(dst, dst_addr, ptl_id, &rva);
         if (!r_mkey) {
             SPML_ERROR("pe=%d: %p is not address of shared variable",
                        dst, dst_addr);
@@ -1129,7 +1135,11 @@ static inline int mca_spml_ikrit_put_internal(void* dst_addr,
         put_req->mxm_req.opcode = MXM_REQ_OP_PUT;
     }
     if (!zcopy) {
-        put_req->mxm_req.flags |= MXM_REQ_SEND_FLAG_BLOCKING;
+        if (size < mca_spml_ikrit.put_zcopy_threshold) {
+            put_req->mxm_req.flags |= MXM_REQ_SEND_FLAG_BLOCKING;
+        } else {
+            put_req->mxm_req.opcode = MXM_REQ_OP_PUT_SYNC;
+        }
     }
 #endif
 
@@ -1159,15 +1169,8 @@ static inline int mca_spml_ikrit_put_internal(void* dst_addr,
 
     mca_spml_ikrit.mxm_peers[dst]->n_active_puts++;
 
-    mxm_req_send(&put_req->mxm_req);
+    SPML_IKRIT_MXM_POST_SEND(put_req->mxm_req);
 
-    if (MXM_OK != put_req->mxm_req.base.error) {
-        OPAL_THREAD_ADD32(&mca_spml_ikrit.n_active_puts, -1);
-        SPML_ERROR("put request %p failed: %s - aborting",
-                   (void*)put_req, mxm_error_string(put_req->mxm_req.base.error));
-        oshmem_shmem_abort(-1);
-        return OSHMEM_ERROR;
-    }
     if (need_progress)
         mxm_progress(mca_spml_ikrit.mxm_context);
 
@@ -1194,11 +1197,7 @@ int mca_spml_ikrit_put_simple(void* dst_addr,
 
     ptl_id = get_ptl_id(dst);
     /* Get rkey of remote PE (dst proc) which must be on memheap  */
-    r_mkey = mca_memheap.memheap_get_cached_mkey(dst,
-                                                 //(unsigned long) dst_addr,
-                                                 dst_addr,
-                                                 ptl_id,
-                                                 &rva);
+    r_mkey = mca_memheap_base_get_cached_mkey(dst, dst_addr, ptl_id, &rva);
     if (!r_mkey) {
         SPML_ERROR("pe=%d: %p is not address of shared variable",
                    dst, dst_addr);
@@ -1221,7 +1220,7 @@ int mca_spml_ikrit_put_simple(void* dst_addr,
         }
         /* segment not mapped - fallback to rmda */
         ptl_id = MXM_PTL_RDMA;
-        r_mkey = mca_memheap.memheap_get_cached_mkey(dst,
+        r_mkey = mca_memheap_base_get_cached_mkey(dst,
                                                      //(unsigned long) dst_addr,
                                                      dst_addr,
                                                      ptl_id,
@@ -1270,13 +1269,7 @@ int mca_spml_ikrit_put_simple(void* dst_addr,
         mca_spml_ikrit.mxm_peers[dst]->need_fence = 1;
     }
 
-    mxm_req_send(&mxm_req);
-    if (MXM_OK != mxm_req.base.error) {
-        SPML_ERROR("put request failed: %s(%d) - aborting",
-                   mxm_error_string(mxm_req.base.error), mxm_req.base.error);
-        oshmem_shmem_abort(-1);
-        return OSHMEM_ERROR;
-    }
+    SPML_IKRIT_MXM_POST_SEND(mxm_req);
 
     wait.req = &mxm_req.base;
     wait.state = (mxm_req_state_t)(MXM_REQ_SENT | MXM_REQ_COMPLETED);
@@ -1363,6 +1356,7 @@ int mca_spml_ikrit_fence(void)
 /* blocking receive */
 int mca_spml_ikrit_recv(void* buf, size_t size, int src)
 {
+    mxm_error_t ret = MXM_OK;
     mxm_recv_req_t req;
     char dummy_buf[1];
 
@@ -1386,9 +1380,12 @@ int mca_spml_ikrit_recv(void* buf, size_t size, int src)
     req.base.data.buffer.length = size == 0 ? sizeof(dummy_buf) : size;
     req.base.data.buffer.memh = NULL;
 
-    mxm_req_recv(&req);
+    ret = mxm_req_recv(&req);
+    if (MXM_OK != ret) {
+        return OSHMEM_ERROR;
+    }
     mca_spml_irkit_req_wait(&req.base);
-    if (req.base.error != MXM_OK) {
+    if (MXM_OK != req.base.error) {
         return OSHMEM_ERROR;
     }
     SPML_VERBOSE(100,
@@ -1429,7 +1426,8 @@ int mca_spml_ikrit_send(void* buf,
     req.base.data.buffer.length = size == 0 ? sizeof(dummy_buf) : size;
     req.base.data.buffer.memh = NULL;
 
-    mxm_req_send(&req);
+    SPML_IKRIT_MXM_POST_SEND(req);
+
     mca_spml_irkit_req_wait(&req.base);
     if (req.base.error != MXM_OK) {
         return OSHMEM_ERROR;
