@@ -12,8 +12,6 @@
 #include <stdarg.h>
 #include <stdio.h>
 
-#include "opal/util/output.h"
-
 #include "oshmem/constants.h"
 #include "oshmem/util/oshmem_util.h"
 
@@ -22,7 +20,7 @@ void oshmem_output_verbose(int level, int output_id, const char* prefix,
 {
     va_list args;
     char *buff, *str;
-    int ret;
+    int ret = 0;
 
     if (level < opal_output_get_verbosity(output_id)) {
         UNREFERENCED_PARAMETER(ret);
@@ -44,6 +42,7 @@ void oshmem_output_verbose(int level, int output_id, const char* prefix,
     }
 }
 
+
 void oshmem_output(int output_id, const char* prefix, const char* file,
     int line, const char* function, const char* format, ...)
 {
